/**
 * This program instruments DSP filters for testing speed and accuracy.
 * Threads on core0:
 * --graphics thread draws the bode plots
 * --serial thread gets frequency sweep range
 * --blinky
 * ISR on core1:
 * -- runs at 40 KHz to generate a sweep sinewave by DDS, then filter it.
 * The filter output is cross correlated with the sine (and cosine) to get the
 * effect on phase and magnitude.
 * -- Without SPI DAC output ISR duration is 3.8 uSec, with the 2-pole IIR taking 2 uSec of that.
 * -- With SPI DAC output the duration is 8 uSec.
 * 
 * Included IIR filters are:
 * 2-pole Butterworth lowpass
 * 2-pole Butterworth bandwpass
 * 2-pole Chebyshev type-1 lowpass
 * 
 * VGA by Hunter Adams (vha3@cornell.edu)
 * converted to 320x240 with 256 colors by Bruce; bruce.land@cornell.edu
 * 
 * HARDWARE CONNECTIONS
 *  - GPIO 16 ---> VGA Hsync
 *  - GPIO 17 ---> VGA Vsync
 * 
 *  - GPIO 08 ---> 330 ohm resistor ---> VGA Blue lo-bit |__ both wired to 150 ohm to ground 
 *  - GPIO 09 ---> 220 ohm resistor ---> VGA Blue hi_bit |   and to VGA blue
 * 
 *  - GPIO 10 ---> 1000 ohm resistor ---> VGA Green lo-bit |__ three wired to 100 ohm to ground
 *  - GPIO 11 ---> 680 ohm resistor ---> VGA Green mid_bit |   and to VGA Green
 *  - GPIO 12 ---> 330 ohm resistor ---> VGA Green hi_bit  |   
 * 
 *  - GPIO 13 ---> 1000 ohm resistor ---> VGA Red lo-bit |__ three wired to 100 ohm to ground
 *  - GPIO 14 ---> 680 ohm resistor ---> VGA Red mid_bit |   and to VGA red
 *  - GPIO 15 ---> 330 ohm resistor ---> VGA Red hi_bit  |   
 * 
 *  - RP2040 GND ---> VGA GND
 *
 * RESOURCES USED
 *  - PIO state machines 0 to 3 on PIO instance 0
 *  - DMA channels 0, 1, 2, 3 data send to two PIO
 *  - 76.8 kBytes of RAM (for pixel color data)
 * color encoding: bits 7:5 red; 4:2 green; 1:0 blue
 *
 * Protothreads v1.1.1
 * graphics dsp plot thread
 * serial thread to set the filter parameters///
 * the usual blinky thread for a hearbeat
 */
// ==========================================
// === VGA graphics library
// ==========================================
#include "vga256_graphics.h"
#include <stdio.h>
#include <stdlib.h>
#include <string.h> 
#include <math.h>
#include "pico/stdlib.h"
#include "hardware/pio.h"
#include "hardware/dma.h"

//===========================================
// === other hardware
//===========================================
#include "hardware/gpio.h"
#include "hardware/timer.h"
#include "hardware/spi.h"
#include "pico/stdlib.h"
#include "hardware/uart.h"
#include "hardware/gpio.h"
#include "hardware/structs/iobank0.h"
#include "hardware/regs/rosc.h"

// ==========================================
// === protothreads globals
// ==========================================
#include "hardware/sync.h"
#include "hardware/timer.h"
#include "pico/multicore.h"
#include "string.h"
#include "math.h"
// protothreads header
#include "pt_cornell_rp2040_v1_1_1.h"
// interactive flag
int new_value = false ;

// ========================================
// === dds setup 
// =======================================
//
// == bit fixed point 2.14 format ===========
// resolution 2^-14 =  6.1035e-5
// dynamic range is +1.9999/-2.0
typedef signed short s1x14 ;
#define muls1x14(a,b) ((s1x14)((((int32_t)(a))*((int32_t)(b)))>>14)) //multiply two fixed s1x14
#define float_to_s1x14(a) ((s1x14)((a)*16384.0)) // 2^14
#define s1x14_to_float(a) ((float)(a)/16384.0)
#define absfix14(a) abs(a)

// =======================================
// 1/Fs in microseconds for ISR
volatile int alarm_period = 25 ;

// DDS variables 
unsigned int main_inc ;
unsigned int main_accum ;
s1x14 sine_table_out, cos_table_out ;
s1x14  filter_out, down_sampled_signal ;
s1x14 sine_table[256], cos_table[256] ;
float Fs=40000, Fout, F_low, F_high, F_inc;

// cross correlation
// phase/amp variables
volatile int I_signal, Q_signal ;
// averaging the signal
int avg_size = 20000;
volatile int avg_count;
// accumulators
volatile int32_t I_avg, Q_avg;
// and the final averages
float I_out, Q_out, mag, phase, I ;//
// modulo countr for downsampling in CIC
 int downsample_clock ;

// ========================================
// === spi setup 
// =======================================
//SPI configurations
#define PIN_CS   5
#define PIN_SCK  6
#define PIN_MOSI 7
#define SPI_PORT spi0

// constant to tell SPI DAC what to do
// prepend to each 12-bit sample
#define DAC_config_chan_A 0b0011000000000000
// B-channel, 1x, active
#define DAC_config_chan_B 0b1011000000000000
uint16_t DAC_data ; 
//
// ==================================================
// === graphics plot bode -- RUNNING on core 0
// ==================================================
        
int start_time, end_time ;
char video_string[32] ;
// graph labels
short decade_name[4] = {10, 100, 1000, 10000};
// short mag_name[5] = {0, -1, -2, -3 , -4, -5};
float mag_name[5] = {1, .1, .01, .001, .0001};

static PT_THREAD (protothread_graphics(struct pt *pt)) {
    PT_BEGIN(pt);
    // the protothreads interval timer
    PT_INTERVAL_INIT() ;

    // background
    fillRect(0, 0, 319, 239, BLACK); // 

    // Draw some filled rectangles
    fillRect(0, 0, 76, 10, BLUE); // blue box
    fillRect(100, 0, 150, 10, GREEN); // red box
    //fillRect(200, 0, 76, 10, GREEN); // green box

    // Write some text
    setTextColor(WHITE) ;
    setCursor(10, 1) ;
    setTextSize(1) ;
    writeString("ECE 4760") ;

    setTextColor(BLACK) ;
    setCursor(102, 1) ;
    setTextSize(1) ;
    writeString("DSP Bode Plots s1x14") ;

    // graphics constants in pixels
    // for plots
    #define decade_width 75 // pixels
    #define decade_start 20
    #define zero_phase 170 // y-cood
    #define gain_phase 0.2 // pixels/degree
    #define zero_mag 25 // y cood 0 db
    #define gain_mag 20 // pixels/decade
    
    while(true) {

        // plot new curve
        if(new_value == true){  
          // set up the graphics
          fillRect(0, 20, 319, 239, BLACK);  
          drawHLine(10, 120, 300, rgb(1,1,1));
          drawHLine(10, 230, 300, rgb(1,1,1));
          // graph Frequency lines
          for (int j=0; j<4; j++){
            drawVLine(20+j*decade_width, 20, 220, rgb(1,1,1));
            drawVLine(20+j*decade_width+23, 20, 210, rgb(1,1,1));
            drawVLine(20+j*decade_width+52, 20, 210, rgb(1,1,1));
            setTextColor(GRAY2) ;
            setCursor(10+j*decade_width, 231) ;
            setTextSize(1) ;
            sprintf(video_string, "%-5d", decade_name[j]) ;
            writeString(video_string) ;
          }
          // mag lines
          for (int j=0; j<5; j++){
            drawHLine(20, zero_mag+j*gain_mag, 280, rgb(1,1,1));
            setTextColor(rgb(0,5,0)) ;
            setCursor(10, zero_mag+j*gain_mag) ;
            setTextSize(1) ;
            sprintf(video_string, "%-5.4f", mag_name[j]) ;
            writeString(video_string) ;
          }
          // phase lines
          // mag lines
          drawHLine(20, zero_phase, 280, rgb(1,1,1));
          setTextColor(rgb(0,5,2)) ;
          setCursor(10, zero_phase) ;
          setTextSize(1) ;
          writeString("O") ;
          drawHLine(20, zero_phase+90*gain_phase, 280, rgb(1,1,1));
          drawHLine(20, zero_phase+180*gain_phase, 280, rgb(1,1,1));
          setCursor(10, zero_phase+180*gain_phase) ;
          writeString("-18O") ;
          drawHLine(20, zero_phase-90*gain_phase, 280, rgb(1,1,1));
          drawHLine(20, zero_phase-180*gain_phase, 280, rgb(1,1,1));
          setCursor(10, zero_phase-180*gain_phase) ;
          writeString("+18O") ;

          // now set up the DSP
            Fout = F_low ;
            printf("freq,   mag,  phase\n\r") ;
            //
            static int i ;
            while(Fout <= F_high){
              // set dds
                main_inc = Fout * pow(2,32 )/ Fs ;
                // the measurement
                I_avg = 0 ;
                Q_avg = 0 ;
                avg_count = 0;
                // wait for # of samples
                while (avg_count<avg_size) {}; 
                // and get the average
                I_out = s1x14_to_float(I_avg / avg_size );
                Q_out = s1x14_to_float(Q_avg / avg_size );

                mag = 2 * sqrt((I_out)*(I_out) + (Q_out)*(Q_out)) ;
                phase = 57.29 * atan2(Q_out,I_out) ;
                printf("%5.0f  %5.3f  %5.1f\n\r", Fout, mag, phase) ;
                // log variables to plot
                float log_mag, log_freq;
                log_mag = -gain_mag * log10(mag) + zero_mag ;
                log_freq = decade_width * (log10(Fout)-1) + decade_start ;
                // magnitude bode plot
                drawPixel(log_freq, log_mag, GREEN) ;
                // phase plot
                drawPixel(log_freq, -gain_phase*phase+zero_phase, CYAN) ;

                // the next frequency on log scale
                Fout = Fout * F_inc;

                PT_YIELD_INTERVAL(10) ;
            }
            new_value = false ;
        }
        main_inc = 0;

        PT_YIELD_INTERVAL(20000); 
   }
   PT_END(pt);
} // graphics thread

// ==================================================
// === toggle25 thread on core 0
// ==================================================
// the on-board LED blinks
static PT_THREAD (protothread_toggle25(struct pt *pt))
{
    PT_BEGIN(pt);
    static bool LED_state = false ;
    
     // set up LED p25 to blink
     gpio_init(25) ;	
     gpio_set_dir(25, GPIO_OUT) ;
     gpio_put(25, true);
     // data structure for interval timer
     PT_INTERVAL_INIT() ;

      while(1) {
        // yield time 0.1 second
        //PT_YIELD_usec(100000) ;
        PT_YIELD_INTERVAL(100000) ;

        // toggle the LED on PICO
        LED_state = LED_state? false : true ;
        gpio_put(25, LED_state);

        // NEVER exit while
      } // END WHILE(1)
  PT_END(pt);
} // blink thread


// ==================================================
// === user's serial input thread on core 0
// ==================================================
// serial_read an serial_write do not block any thread
// except this one


static PT_THREAD (protothread_serial(struct pt *pt))
{
    PT_BEGIN(pt);
        char video_buffer[20];
      //
      while(1) {
        PT_YIELD_usec(100000);
        PT_YIELD_UNTIL(pt, new_value==false) ;

        // print prompt
        sprintf(pt_serial_out_buffer, "F_low, F_high, F_inc: ");
        // spawn a thread to do the non-blocking write
        serial_write ;

        // spawn a thread to do the non-blocking serial read
         serial_read ;
        sscanf(pt_serial_in_buffer, "%f %f  %f", &F_low, &F_high, &F_inc) ;

        // some error checking
        if(F_low < 10) F_low = 10 ;
        if(F_high > Fs/4) F_high = Fs/4 ;
        if(F_inc <= 1.01) F_inc = 1.1 ;
        // 
        new_value = true ;
        // NEVER exit while
      } // END WHILE(1)
  PT_END(pt);
} // serial thread

// ==========================================
// === set up timer ISR  used in this pgm
// ==========================================
// === timer alarm ========================
// !! modifiying alarm zero trashes the cpu 
//        and causes LED  4 long - 4 short
// !! DO NOT USE alarm 0
// This low-level setup is ocnsiderably faster to execute
// than the hogh-level callback

#define ALARM_NUM 1
#define ALARM_IRQ TIMER_IRQ_1
// ISR interval will be 10 uSec
//
// the actual ISR
void compute_sample(void);
//
static void alarm_irq(void) {
    // mark ISR entry
    gpio_put(2,1);
    // Clear the alarm irq
    hw_clear_bits(&timer_hw->intr, 1u << ALARM_NUM);
    // arm the next interrupt
    // Write the lower 32 bits of the target time to the alarm to arm it
    timer_hw->alarm[ALARM_NUM] = timer_hw->timerawl + alarm_period ;
    //
    compute_sample();

    // mark ISR exit
    gpio_put(2,0);
}

// set up the timer alarm ISR
static void alarm_in_us(uint32_t delay_us) {
    // Enable the interrupt for our alarm (the timer outputs 4 alarm irqs)
    hw_set_bits(&timer_hw->inte, 1u << ALARM_NUM);
    // Set irq handler for alarm irq
    irq_set_exclusive_handler(ALARM_IRQ, alarm_irq);
    // Enable the alarm irq
    irq_set_enabled(ALARM_IRQ, true);
    // Enable interrupt in block and at processor
    // Alarm is only 32 bits 
    uint64_t target = timer_hw->timerawl + delay_us;
    // Write the lower 32 bits of the target time to the alarm which
    // will arm it
    timer_hw->alarm[ALARM_NUM] = (uint32_t) target;   
}

//======================================================== 
// Second order Butterworth lowpass
// xx is the current input signal sample
// returns the current filtered output sample
// Simplification: b(1)*x(n)+b(2)*x(n-1)+b(3)*x(n-2) = 
// b(1)*x(n)+2*b(1)*x(n-1)+b(1)*x(n-2) =
// b(1)* (x(n)+(x(n-1)<<1)+x(n-2))
s1x14 Butter2low(s1x14 xx){
  static s1x14 yy  ; 
  //IIR state variables
  static s1x14  b1,  a2, a3 ;
  static s1x14  xn, xn_1, xn_2 ; 
  static s1x14  yn_1, yn_2  ;

  //2-pole butterworth lowpass, cutoff=500/20000 , gain=1 
  // for 40 khz sample Nyquist is 20000 Hz,
  b1 = 24 ;             
  a2 = (-30950) ;   
  a3 = (14662) ;  

  // sum the 5 terms: yy += xx*coeff 
  // and update the state variables
  // as soon as possible

  yy = muls1x14((xx+(xn_1<<1)+xn_2), b1);
  xn_2 = xn_1;
  xn_1 = xx;
  yy = yy - muls1x14(yn_2, a3); 
  yn_2 = yn_1;
  yy = yy - muls1x14(yn_1, a2); 
  yn_1 = yy;
  return yy;
} 

//======================================================== 
// Second order cheby1 lowpass
// NOTE: this is the compensator for the downsampler and
// should NOT be changed!
// =======================================================
// xx is the current input signal sample
// returns the current filtered output sample
// Simplification: b(1)*x(n)+b(2)*x(n-1)+b(3)*x(n-2) = 
// b(1)*x(n)+2*b(1)*x(n-1)+b(1)*x(n-2) =
// b(1)* (x(n)+(x(n-1)<<1)+x(n-2))
  s1x14 ChebyOne2low(s1x14 xx){
  static s1x14 yy  ; 
  //IIR state variables
  static s1x14  b1,  a2, a3 ;
  static s1x14  xn, xn_1, xn_2 ; 
  static s1x14  yn_1, yn_2  ;

  // CIC compensator
  // [b,a] = cheby1(2, 8, 0.5)
  b1 = 1927  ;            
  a2 = -8079  ;  
  a3 = 11053 ;  

  // sum the 5 terms: yy += xx*coeff 
  // and update the state variables
  // as soon as possible

  yy = muls1x14((xx+(xn_1<<1)+xn_2), b1);
  xn_2 = xn_1;
  xn_1 = xx;
  yy = yy - muls1x14(yn_2, a3); 
  yn_2 = yn_1;
  yy = yy - muls1x14(yn_1, a2); 
  yn_1 = yy;
  return yy;
} 

//======================================================== 
// Second order Butterworth bandpass
// xx is the current input signal sample
// b(1)*x(n)+b(2)*x(n-1)+b(3)*x(n-2) = 
// b(1)*x(n)+0*b(1)*x(n-1)-b(1)*x(n-2) =
// b(1)* (x(n)-x(n-2))
s1x14 Butter2band(s1x14 xx)
{
  static s1x14 yy ; 
  //IIR state variables
  static s1x14  b1,  a2, a3 ;
  static s1x14  xn, xn_1, xn_2 ; 
  static s1x14  yn_1, yn_2  ;	

  //  [b,a]= butter(1, [200/20000 500/20000])
  // b = 0.023024  0  -0.023024
  // a = 1.0000  -1.9515   0.9540
  b1 = float_to_s1x14(0.02302)  ;             
  a2 = float_to_s1x14(-1.9515) ;  //s7x24() ;   
  a3 = float_to_s1x14(0.9540) ; // //s7x24() ;  

	yy = muls1x14((xx-xn_2), b1);
	xn_2 = xn_1;
	xn_1 = xx;
	yy = yy - muls1x14(yn_2,a3); 
	yn_2 = yn_1;
	yy = yy - muls1x14(yn_1,a2); 
    yn_1 = yy;
    return yy;
} 

//========================================================   
// 4th order CIC
// / CIC filter
// N=4, R=4, M=2   
// -- N=order=4 (4 integrators)
// -- R=downsample=4 -- M=delay(after downsample)=2
// output rate is 10 KHz sample rate

s1x14 CIC_n4_r4_m2(s1x14 xx)
{
     //CIC state variables
 static int32_t integrator1, integrator2, integrator3, integrator4 ;
 static int32_t  last_integrator4, last2_integrator4 ;
 static int32_t  last_integrator2, last2_integrator2 ;
 static int32_t  comb1, comb2, comb3, comb4 ;
 static int32_t  last_comb1, last_comb2, last_comb3, last_comb4 ; 
 static int32_t  last2_comb1, last2_comb2, last2_comb3, last2_comb4 ; 
 //https://dsp.stackexchange.com/questions/19584/how-to-make-cic-compensation-filter
 static s1x14 yy, zz ;

     integrator1 = integrator1 + xx ;
	 integrator2 = integrator2 + integrator1 ;
	 integrator3 = integrator3 + integrator2 ;
	 integrator4 = integrator4 + integrator3 ;
	 
	 if ((downsample_clock & 0x03) == 0) //modulo 4
	 {
		comb1 = integrator4 - last2_integrator4 ;
    	comb2 = comb1 - last2_comb1 ;
    	comb3 = comb2 - last2_comb2 ;
        //
    	yy = (s1x14)((comb3 - last2_comb3)>>11) ; //16
		
		// update state
    	last2_integrator4 = last_integrator4 ;
    	last_integrator4 = integrator4 ;
    	last2_comb1 = last_comb1 ;
    	last_comb1 = comb1;
    	last2_comb2 = last_comb2 ;
    	last_comb2 = comb2;
    	last2_comb3 = last_comb3 ;
    	last_comb3 = comb3;
        // this is a compensator to flatten downsample spectrum
        // increases bandwidth from 0.2 to 0.4 of NEW Nyquist freq
        // [b,a] = cheby1(2, 8, 0.5)
        zz = ChebyOne2low(yy) ;
	 }
    return zz;
} 

// ==================================================
// === dsp ISR -- RUNNING on core 1 -- 40 KHz
// Nyquist frequency is 20 KHz
// ==================================================
// 
void compute_sample(void)
{
    // make the sine wave to send to filter
    main_accum += main_inc ;
    // update main waveform
    sine_table_out = sine_table[main_accum>>24] ;
    cos_table_out = cos_table[main_accum>>24] ;

    // convert to int, shift to all positive and mask to 12 bits
    // UNCOMMNET the next two lies for SPI output
    //DAC_data = (DAC_config_chan_B | (((sine_table_out>>14) + 2048) & 0xfff))  ;
    //spi_write16_blocking(SPI_PORT, &DAC_data, 1) ;

    // Now the filter
    // with  timing by gpio pin
    gpio_put(2,0);
    // down sampled signal updated every 4th ISR
    down_sampled_signal =  CIC_n4_r4_m2(sine_table_out) ;
    // all four phases of the downsampler can be used for different
    // filters: phase 0 is used by the downsample process itself
    if((downsample_clock & 0x03) == 1){
        filter_out = Butter2low(down_sampled_signal);
    }
    gpio_put(2,1);

     // uncomment for outpput
     // shifting right 14 bits converts s7x24 to a 10 bit number for the DAC
     // (assumes unity gain for filters)
     // DAC_data = (DAC_config_chan_A | (((filter_out>>14) + 2048) & 0xfff))  ; //14
     // DAC_data = (DAC_config_chan_A | (((downsample_clock & 0x03)<<9) & 0xfff))  ;
     // spi_write16_blocking(SPI_PORT, &DAC_data, 1) ;
      
    // modulo counter for downsampling
    // by 4, so new Nyquist is 5 KHz for decimaed filters
    downsample_clock++ ;

    // do quadrature multiplies for amp and phase
    I_signal = muls1x14(filter_out, sine_table_out );
    Q_signal = muls1x14(filter_out, cos_table_out );
    // accumulate signal for averaging
    // main thread waits for count to reach maximum value
    if (avg_count<avg_size) {
        I_avg += I_signal ;
        Q_avg += Q_signal ;
        avg_count++ ;
    }
} // isr sample routine

// ========================================
// === core 1 main -- started in main below
// ========================================
void core1_main(){ 
  //
  //  === add threads  ====================
  // for core 1
  // testing timing
    gpio_init(2) ;	
    gpio_set_dir(2, GPIO_OUT) ;
  // fire off interrupt
    alarm_in_us(alarm_period);
    while(1) {};
  //
  // === initalize the scheduler ==========
  //pt_schedule_start ;
  // NEVER exits
  // ======================================
}

// ========================================
// === core 0 main
// ========================================
int main(){
  // set the clock
  //set_sys_clock_khz(250000, true); // 171us
  // start the serial i/o
  stdio_init_all() ;
  // announce the threader version on system reset
  printf("\n\rProtothreads RP2040 v1.11 two-core\n\r");

  // Initialize the VGA screen
  initVGA() ;

  // Initialize SPI channel (channel, baud rate set to 20MHz)
    // connected to spi DAC
    spi_init(SPI_PORT, 20000000) ;
    // Format (channel, data bits per transfer, polarity, phase, order)
    spi_set_format(SPI_PORT, 16, 0, 0, 0);
    // Map SPI signals to GPIO ports
    //gpio_set_function(PIN_MISO, GPIO_FUNC_SPI);
    gpio_set_function(PIN_SCK, GPIO_FUNC_SPI);
    gpio_set_function(PIN_MOSI, GPIO_FUNC_SPI);
    gpio_set_function(PIN_CS, GPIO_FUNC_SPI) ;

    int i;
    for (i=0; i<256; i++) {
      // sine table is in natural +1/-1 range
      sine_table[i] = float_to_s1x14(sin(2*3.1416*(float)i/256)) ;
      cos_table[i] = float_to_s1x14(cos(2*3.1416*(float)i/256)) ;
    }
   // printf("%f\n\r", s15x16_to_float(sine_table[64])) ;

  // start core 1 threads
  multicore_reset_core1();
  multicore_launch_core1(&core1_main);

  // === config threads ========================
  // for core 0
  pt_add_thread(protothread_graphics);
  pt_add_thread(protothread_toggle25);
  pt_add_thread(protothread_serial) ;
  //
  // === initalize the scheduler ===============
  pt_schedule_start ;
  // NEVER exits
  // ===========================================
} // end main